//////////////////////////////////////////////////////////////////////////////////
// Stuff (c) 2007 G-Truc Creation (www.g-truc.net)
//////////////////////////////////////////////////////////////////////////////////
// File    : async_array.inl
// Author  : Christophe [Groove] Riccio (g.truc.creation [at] gmail [point] com)
// Created : 2007-08-02
// Updated : 2007-08-02
// Licence : This source is under GNU LGPL licence
//////////////////////////////////////////////////////////////////////////////////

namespace stuff
{	
	template<typename T, class D>
	async_array<T, D>::async_array():
		object(new T*(0)),
		counter(new size_type(1))
	{}

	template<typename T, class D>
	async_array<T, D>::async_array(bool Null) :
		object(Null ? 0 : new T*(0)),
		counter(Null ? 0 : new size_type(1))
	{}

	template<typename T, class D>
	async_array<T, D>::async_array(const async_array<T, D>& r) :
		object(r.object),
		counter(r.counter),
		deleter(r.deleter)
	{
		if(counter != 0)
			++(*counter);
	}

	template<typename T, class D>
	async_array<T, D>::async_array(T* p):
		object(new T*(p)),
		counter(new size_type(1))
	{}

	template<typename T, class D>
	async_array<T, D>::async_array(T* p, D d):
		object(new T*(p)),
		counter(new size_type(1)),
		deleter(d)
	{}

	template<typename T, class D>
	async_array<T, D>::~async_array()
	{
		_release();
	}

	template<typename T, class D>
	void async_array<T, D>::_release()
	{
		if(counter == 0 && object != 0)
		{
			delete object;
			object = 0;
		}

		if(counter != 0)
		{
			(*counter)--;
			if(*counter <= 0)
			{
				_free();
				delete object;
				object = 0;
				delete counter;
				counter = 0;
			}
		}
	}

	template<typename T, class D>
	void async_array<T, D>::_free()
	{
		if(object == 0)
			return;

		if(*object != 0)
			deleter(*object);
		*object = 0;
	}

	template<typename T, class D>
	bool async_array<T, D>::operator==(const async_array<T, D>& r) const
	{
		return counter == r.counter;
	}

	template<typename T, class D>
	bool async_array<T, D>::operator!=(const async_array<T, D>& r) const
	{
		return counter != r.counter;
	}

	template<typename T, class D>
	bool async_array<T, D>::operator<(const async_array<T, D>& r) const
	{
		return counter < r.counter;
	}

	template<typename T, class D>
	bool async_array<T, D>::operator==(const int Null) const
	{
		if(Null != 0) return false;
		return counter == 0;
	}

	template<typename T, class D>
	bool async_array<T, D>::operator!=(const int Null) const
	{
		if(Null != 0) return false;
		return counter != 0;
	}

	template<typename T, class D>
	bool async_array<T, D>::operator!() const
	{
		return counter == 0;
	}

	template<typename T, class D>
	async_array<T, D>& async_array<T, D>::operator=(const async_array<T, D>& r)
	{
		_release();
		this->object = r.object;
		this->counter = r.counter;
		this->deleter = r.deleter;
		if(this->counter != 0)
			(*this->counter)++;
		return *this;
	}

	template<typename T, class D>
	inline T& async_array<T, D>::operator[](std::ptrdiff_t i) const
	{
		//BOOST_ASSERT(object != 0);
		//BOOST_ASSERT(i >= 0);
		return (*object)[i];
	}

/*
	template<typename T, class D>
	void async_array<T, D>::reset()
	{
		async_array<T, D>().swap(*this);
	}

	template<typename T, class D>
	void async_array<T, D>::reset(T* p)
	{
		async_array<T, D>(p).swap(*this);
	}
*/

	template<typename T, class D>
	void async_array<T, D>::reset()
	{
		_free();
	}

	template<typename T, class D>
	void async_array<T, D>::reset(T* p)
	{
		if(counter == 0)
		{
			_release();
			*this = async_array<T, D>(p);
		}
		else
		{
			if(object == 0)
				object = new T*;

			_free();

			*object = p;
		}
	}

	template<typename T, class D>
	void async_array<T, D>::reset(T* p, D d)
	{
		async_array<T, D>(p, d).swap(*this);
	}

	template<typename T, class D>
	T* async_array<T, D>::get() const
	{
		if(object == 0)
			return 0;
		return *object;
	}

	template<typename T, class D>
	T* async_array<T, D>::release()
	{
		if(object == 0)
			return 0;

		T* result = *object;
		*object = 0;
		return result;
	}

	template<typename T, class D>
	void async_array<T, D>::swap(async_array<T, D>& r)
	{
		if(object == 0 || r.object == 0)
			return;

		T* tmp = *(r.object);
		*(r.object) = *object;
		*object = tmp;
	}

	template<typename T, class D>
	typename async_array<T, D>::size_type async_array<T, D>::use_count() const
	{
		if(counter == 0)
			return 0;
		return *counter;
	}

	template<typename T, class D>
	bool async_array<T, D>::empty() const
	{
		if(object == 0)
			return true;
		return *object == 0;
	}

	template<typename T, class D>
	bool async_array<T, D>::unique() const
	{
		if(counter == 0)
			return 0;
		return *counter == 1;
	}

	template<typename T, class D>
	const async_array<T, D> async_array<T, D>::null(true);
}
